"""
Helper for building a simple thread.
"""

from __future__ import annotations

from math import pi, tan

try:
    import bd_warehouse.thread as _t
    import cadquery
except ImportError:
    _t = None
    TrapezoidalThread = object
else:
    TrapezoidalThread = _t.TrapezoidalThread

__all__ = ["AngledThread", "ISO228_Thread"]

IN = 25.4  # mm per inch


def radians(x):
    return x * pi / 180


def AngledThread(
    radius,
    offset,
    apex=0,
    angle=45,
    external: bool = True,
    simple: bool = False,
    **kw,
):
    """
    Create a thread with a defined angle.

    The thread is always an add-on object. Thus, you should add
    internal threads to a hole with radius (@radius + @offset - 𝛆).

    @radius: the radius of the hole.
    @offset: the width of the thread.
    @pitch: distance between windings
    @length: length of screw / hole
    @apex: flat area on top (default zero)
    @angle: slope angle (default 45°).
    @external: screw or nut? (default True=screw)
    @hand: "left" or "right" (default "right")
    @end_finishes: 2-tuple of ["raw", "square", "fade", "chamfer"]
    """
    # TODO support tapered screws

    if external:
        apex_radius = radius + offset
    else:
        apex_radius = radius
        radius += offset
    root_width = tan(angle * pi / 180) * offset * 2 + apex

    fin = kw.get("end_finishes")
    if fin is not None and isinstance(fin, str):
        kw["end_finishes"] = fin.split(",")

    return _t.Thread(
        apex_radius=apex_radius,
        apex_width=0.1,
        root_radius=radius,
        root_width=root_width - 0.1,
        simple=simple,
        **kw,
    )


class ISO228_Thread(TrapezoidalThread):
    "Threads for fittings. ISO 228."

    specs = {
        # nominal size: threads per inch / diameter at center of thread
        "1/16": (28, 7.142),
        "1/8": (28, 9.147),
        "1/4": (19, 12.301),
        "3/8": (19, 15.806),
        "1/2": (14, 19.793),
        "5/8": (14, 21.749),
        "3/4": (14, 25.279),
        "7/8": (14, 29.039),
        "1": (11, 31.770),
        "1 1/8": (11, 36.418),
        "1 1/4": (11, 40.431),
        "1 1/2": (11, 46.324),
        "1 3/4": (11, 52.267),
        "2": (11, 58.135),
        "2 1/4": (11, 64.231),
        "2 1/2": (11, 73.705),
        "2 3/4": (11, 80.055),
        "3": (11, 86.405),
        "3 1/2": (11, 98.851),
        "4": (11, 111.551),
        "4 1/2": (11, 124.251),
        "5": (11, 136.951),
        "5 1/2": (11, 149.651),
        "6": (11, 162.351),
    }

    thread_angle = 27.5  # degrees

    @classmethod
    def parse_size(cls, x):
        "not called here. Yes we're duck typing."
        raise RuntimeError("Not applicable")

    def __init__(
        self,
        size: str,
        length: float,
        adj: float = 0,
        external: bool = True,
        **kw,
    ):
        kw.setdefault("end_finishes", ("fade", "fade"))

        self.size = size
        self.external = external
        self.length = length
        (pitch, diameter) = self.specs[self.size]
        self.pitch = IN / pitch
        diameter += 2 * adj

        self.adj = adj

        # 1/6th of the total height, given the thread angle, is cut off.
        # This corresponds to 1/6th of the total width, i.e. the pitch.

        top = 1 / 6 * self.pitch
        apex_width = top
        root_width = self.pitch - top  # - bottom really, but top==bottom here

        # The height difference between the thread's center and its
        # adjacent top (or bottom) is spread over 1/6th of the pitch
        # length; the ratio is defined by the angle.

        hd = self.pitch / 6 / tan(radians(self.thread_angle))
        if self.external:
            self.apex_radius = diameter / 2 + hd
            self.root_radius = diameter / 2 - hd
            self.diameter = 2 * self.root_radius
        else:
            self.apex_radius = diameter / 2 - hd
            self.root_radius = diameter / 2 + hd
            self.diameter = 2 * self.apex_radius

        cq_object = _t.Thread(
            apex_radius=self.apex_radius,
            apex_width=apex_width,
            root_radius=self.root_radius,
            root_width=root_width,
            pitch=self.pitch,
            length=self.length,
            **kw,
        )
        self.end_finishes = cq_object.end_finishes
        self.hand = "right" if cq_object.right_hand else "left"
        cadquery.Solid.__init__(self, cq_object.wrapped)



# reimplementation of basic gridfinity box design by Zack Freedman
# this revision is dated 2022-05-18
# designed in CadQuery and CQ-Editor both checked out from github on 2022-04-13

# Source: @jdegs
# https://www.printables.com/model/174715-basic-undivided-gridfinity-boxes-cadquery-customiz/files
# License: CC 4.0 Attribution
from __future__ import annotations

import cadquery as cq
import math


def gridbox(x_grid_number=2, y_grid_number=2, unit_height=6, disable_mholes=True):
    # adjust these three variables to change the grid size

    # box stackability is implemented!
    # boxes are always 0.5mm shy of the unit_width*42mm, and
    # that is independent of number of units

    # box interior is very minimalistic in this version

    ############################################
    # shouldn't need to adjust anything below here
    box_wd = 42.0  # mm, distance between major straight walls
    box_maj_wd = box_wd  # mm, x,y width of box to add
    box_to_box_clear = 0.5  # mm, applied only once to each x and y
    c_rad = 4  # mm, corner radius of outer box
    socket_ht = 5  # mm, overall socket height
    ov_ht = 3.8 + 7 * unit_height  # mm, overall height for a Z-unit box
    vw_ht = ov_ht - socket_ht  # mm, height of vertical walls
    wall_th = 1.0  # mm, wall thickness, don't adjust as it will break top-socket compat

    # for magnet/bolt holes
    mag_diam = 6.5  # mm
    mag_dep = 2.4  # mm
    bolt_diam = 3.0  # mm
    bolt_dep = 3.6 + mag_dep  # mm, total depth of entire hole includng 3.6mm for the bolt itself
    mag_dist0 = 26 / 2  # mm, distance from holes to zero-axis (26mm between holes total)

    bot_th = 2.0 + bolt_dep - socket_ht  # mm, bottom thickness of box interior
    # bot_th was revised to account for bolt_dep since it is longer than socket_ht
    # the revision ensures that there is 2.0mm of bottom thickness above the top of the bolt holes
    bot_fillet = 1  # mm, radius of fillet on the inside of box bottom

    # actual hypotenuse of t_chm is 3.39mm
    # actual exterior hypotenuse of t_chm is 3.04mm due to cropping

    # for socket only
    b_chm_ht = 1.13 / math.sqrt(2)  # mm, base chamfer height
    strt_ht = 1.8  # mm, straight wall height
    t_chm_ht = socket_ht - b_chm_ht - strt_ht  # top chamfer height

    # create 2D sketch with rounded corners, for the bottom of the socket
    s3a = (
        cq.Sketch()
        .rect(box_wd + 0.001, box_wd + 0.001)  # ensure boxes overlap with small add
        .vertices()
        .fillet(c_rad)
    )

    # take s3a sketch and create one socket for the bottom of box
    f2 = (
        cq.Workplane("XY")
        .placeSketch(s3a)
        .extrude(t_chm_ht * math.sqrt(2), taper=45)
        .faces(">Z")
        .wires()
        .toPending()
        .extrude(strt_ht)
        .faces(">Z")
        .wires()
        .toPending()
        .extrude(b_chm_ht * math.sqrt(2), taper=45)
        .mirror(mirrorPlane="XY")  # flip upside down
    )
    return f2

    # create sketch for exterior box walls
    box_wd_xwall = box_wd * x_grid_number - box_to_box_clear  # 0.5mm for clearance btwn boxes
    box_wd_ywall = box_wd * y_grid_number - box_to_box_clear

    s4 = cq.Sketch().rect(box_wd_xwall, box_wd_ywall).vertices().fillet(c_rad)

    # define x,y positions of the exterior walls
    wall_xpos = (box_wd) * (x_grid_number - 1) / 2
    wall_ypos = (box_wd) * (y_grid_number - 1) / 2

    # create exterior box walls
    f7 = (
        cq.Workplane("XY")
        .placeSketch(s4)
        .extrude(vw_ht)  # vertical wall height
        .translate((wall_xpos, wall_ypos, 0))
    )

    # create tool to trim off excess socket underhang
    f11 = (
        cq.Workplane("XY")
        .placeSketch(s4)
        .extrude(-socket_ht - 1)  # socket is below XY-plane
        .translate((wall_xpos, wall_ypos, 1 / 2))
    )

    c_rad_in = c_rad - wall_th  # mm, corner radius of inner box (f(wall_th))

    # create sketch for box interior that is later swept straight/chamfered
    s5 = (
        cq.Sketch()
        # should NOT be a f(wall_th) to ensure compatibility, but could create other issues
        .rect(box_wd_xwall - 2 * wall_th, box_wd_ywall - 2 * wall_th)
        .vertices()
        .fillet(c_rad_in)
    )

    # for top socket of box walls that allows stacking
    # ordered vars from lowest to highest
    wall_strt_ht1 = vw_ht - 6.5 - bot_th  # main straight interior
    wall_chm_ht1 = 2.33 / math.sqrt(2)  # 45-deg section to support socket w/o supports
    wall_strt_ht2 = 1.2
    wall_chm_ht2 = 0.98 / math.sqrt(2)
    wall_strt_ht3 = 1.8
    wall_chm_ht3 = (
        1.59 / math.sqrt(2) + 0.1
    )  # add a little extra to make sure it clears the top face

    # create tool to later subtract inside of box
    # includes fillets on the bottom interior
    # and chamfered walls that enable stacking
    f9 = (
        cq.Workplane("XY")
        .placeSketch(s5)
        .extrude(wall_strt_ht1)  # vertical wall height
        .translate((wall_xpos, wall_ypos, bot_th))  # REVISED, to make sure bot_th is correct
        .faces("<Z")
        .edges("|X or |Y")
        .fillet(bot_fillet)  # fillet bottom interior
        .faces(">Z")
        .wires()
        .toPending()
        .extrude(wall_chm_ht1 * math.sqrt(2), taper=45)  # "overhang" chamfer"
        .faces(">Z")
        .wires()
        .toPending()
        .extrude(wall_strt_ht2)  # short straight section
        .faces(">Z")
        .wires()
        .toPending()
        .extrude(wall_chm_ht2 * math.sqrt(2), taper=-45)  # 1st part of socket interface
        .faces(">Z")
        .wires()
        .toPending()
        .extrude(wall_strt_ht3)  # next short straight section for socket interface
        .faces(">Z")
        .wires()
        .toPending()
        .extrude(wall_chm_ht3 * math.sqrt(2), taper=-45)  # last part of socket interface
    )

    # pts is the locations of each grid position for each box
    # only used for socket base(s)
    pts = [
        (x * box_maj_wd, y * box_maj_wd)
        for x in range(x_grid_number)
        for y in range(y_grid_number)
    ]
    pts.pop(0)  # remove first element as not to duplicate existing socket
    f2a = f2.pushPoints(pts).eachpoint(
        lambda loc: f2.val().moved(loc),
        combine="a",
    )  # join all sockets
    f4 = f11.intersect(f2a)  # trim excess socket overhang, approx 0.25mm all sides
    f3 = f7.union(f4)  # join trimmed sockets with walls
    f8 = f3.cut(f9)  # remove box interior

    # hole_pts is center locations of ALL the magnet/bolt holes
    # BEWARE that the coordinate system is inverted in the y-direction hence the -1*
    # this is why you should not use cboreHole, because it cant do inverted holes!
    # "counterbore" all the magnet/bolt hole locations
    if not disable_mholes:
        hole_pts = [
            (
                (x * box_maj_wd - mag_dist0 * (1 - 2 * i)),
                -1 * (y * box_maj_wd - mag_dist0 * (1 - 2 * j)),
            )
            for x in range(x_grid_number)
            for y in range(y_grid_number)
            for i in [0, 1]
            for j in [0, 1]
        ]

        f8 = (
            f8.faces("<Z")
            .workplane()
            .pushPoints(hole_pts)
            .cboreHole(bolt_diam, mag_diam, mag_dep, depth=bolt_dep)
        )

    return f8



from __future__ import annotations

try:
    import cadquery as cq
    import numpy as np
except ImportError:
    cq = None


def rotate(v, k, theta):
    "rotate vector @v around axis @k at angle @theta(degrees)"
    v = v.toTuple()
    k = k.toTuple()
    theta *= np.pi / 180

    v = np.asarray(v)
    k = np.asarray(k) / np.linalg.norm(k)  # Normalize k to a unit vector
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    term1 = v * cos_theta
    term2 = np.cross(k, v) * sin_theta
    term3 = k * np.dot(k, v) * (1 - cos_theta)

    return cq.Vector(tuple(term1 + term2 + term3))


def _copy(self):
    "returns a copy of a plane"
    return cq.Plane(origin=self.origin, xDir=self.xDir, normal=self.zDir)


def _translated(self, x, y=None):
    if y is None:
        v = cq.Vector(x[0], x[1], 0)
    else:
        v = cq.Vector(x, y, 0)
    w = self.newObject(self.all())
    w.plane = self.plane.copy()
    w.plane.setOrigin2d(v.x, v.y)
    return w


def _rotated(self, theta, center=None):
    "return a rotated workspace"
    p = self.plane
    w = self.newObject(self.all())
    if center is not None:
        raise NotImplementedError("Math")
    p = p.rotated((0, 0, theta))
    # cq.Plane(origin=p.origin,xDir=rotate(p.xDir,p.normal,theta),normal=p.normal)
    w.plane = p
    return w


def _wp(self, *a, **k):
    return self.copyWorkplane(cq.Workplane(*a, **k))


def _at(self, x, y=None, z=0):
    "move the Z offset"
    if y is not None:
        x = cq.Vector(x, y, z)
    return self.pushPoints((x,))


def _rot(self, angle, end):
    return self.rotate((0, 0, 0), end, angle)


if cq is not None:
    cq.Plane.copy = _copy
    cq.Workplane.translated = _translated
    cq.Workplane.rotated = _rotated
    cq.Workplane.at = _at
    cq.Workplane.wp = _wp
    cq.Workplane.rot_x = lambda s, a: _rot(s, a, (1, 0, 0))
    cq.Workplane.rot_y = lambda s, a: _rot(s, a, (0, 1, 0))
    cq.Workplane.rot_z = lambda s, a: _rot(s, a, (0, 0, 1))

    cq.Workplane.off_x = lambda s, x: s.translate((x, 0, 0))
    cq.Workplane.off_y = lambda s, y: s.translate((0, y, 0))
    cq.Workplane.off_z = lambda s, z: s.translate((0, 0, z))

    _S = cq.occ_impl.shapes.Shape
    _S.rot_x = lambda s, a: _rot(s, a, (1, 0, 0))
    _S.rot_y = lambda s, a: _rot(s, a, (0, 1, 0))
    _S.rot_z = lambda s, a: _rot(s, a, (0, 0, 1))

    _S.off_x = lambda s, x: s.translate((x, 0, 0))
    _S.off_y = lambda s, y: s.translate((0, y, 0))
    _S.off_z = lambda s, z: s.translate((0, 0, z))



"""
Collection of useful shortcuts for 3D modelling with build123d.
"""

from __future__ import annotations

from build123d import (
    Shape,
    Align,
    Axis,
    BuildLine,
    Rot,
    Matrix,
    Intrinsic,
    JernArc,
    Location,
    RegularPolygon,
    Part,
    Plane,
    export_step,
    export_stl,
    import_step,
    import_stl,
    Cylinder,
    loft,
)
from pathlib import Path
import math
from moat.util import InexactFloat

# from bd_warehouse.gear import InvoluteToothProfile, SpurGear, SpurGearPlan

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collection.abc import Sequence

import logging

log = logging.getLogger("moat.d3")

__all__ = [
    "AMIN",
    "ACCM",
    "Loc",
    "LocR",
    "Cyl",
    "Up",
    "D",
    "ArcAt",
    "RotAxis",
    "RotateTo",
    "PolyCap",
    "sin",
    "cos",
    "tan",
    "atan",
    "stack",
    "show",
    "quarter",
    "gear_turn",
    "read_step",
    "read_stl",
    "pos",
    "li",
    "lto",
    "RX",
    "RX1",
    "RX2",
    "RX3",
    "RY",
    "RY1",
    "RY2",
    "RY3",
    "RZ",
    "RZ1",
    "RZ2",
    "RZ3",
    "R0",
    "R1",
    "R2",
    "R3",
]

AMIN = {"align": (Align.MIN, Align.MIN, Align.MIN)}
ACCM = {"align": (Align.CENTER, Align.CENTER, Align.MIN)}


def D(n, **kw):
    """
    A wrapper for `moat.util.InexactFloat`. Saves on typing.
    """
    return InexactFloat(n, **kw)


def Loc(a, b, c):
    """
    Shorthand for Location.
    """
    return Location((a, b, c))


def LocR(d, a, h=0):
    """
    Locate radially, distance d, angle a, clockwise from top.
    """
    a = a * math.pi / 180
    return Loc(d * math.sin(a), d * math.cos(a), h)


def Cyl(d, h):
    """
    A cylinder with diameter d and height h, centered but with +z.
    """
    return Cylinder(d / 2, h, align=(Align.CENTER, Align.CENTER, Align.MIN))


def Up(z):
    """
    Move something up. Shorthand for Location(0,0,z).
    """
    return Location((0, 0, z))


def ArcAt(line, radius, angle=90, twist=0, at_end=True):
    """
    Add an arc to the end of a path. @twist modifies the direction the arc
    bends towards.
    """
    res = JernArc((0, 0, 0), (0, 1, 0), radius, angle)
    res = RX1 * Rot(0, twist, 0) * res

    if at_end:
        return Location(line @ 1) * RotateTo(res, line % 1)
    else:
        return Location(line @ 0) * RotateTo(res, -(line % 0))

    # Location(line@.9)*Rot(line%1)*JernArc((0,0,0),(1,0,0),radius,angle)
    # Rot(0,0,twist)*


def PolyCap(d, n=6):
    """
    Given a polygonal rod, this returns its end cap.
    """
    al = (Align.CENTER, Align.CENTER)
    return stack(RegularPolygon(d / 2, n, align=al), d / 4, RegularPolygon(d / 4, n, align=al))


def RotateTo(thing, direction):
    ax = Axis((0, 0, 0), (0, 0, 1) + direction.normalized())
    return thing.rotate(ax, 180)


def RotAxis(axis, angle):
    from transforms3d.euler import mat2euler

    rtm = Matrix()
    breakpoint()
    rtm.rotate(axis, angle)
    rtm = [[rtm[x, y] for y in range(3)] for x in range(3)]
    rtm = [a * 180 / math.pi for a in mat2euler(rtm, "szyx")]
    return Rot(*rtm, Intrinsic.XYZ)


def sin(a):
    "sin (angle in degrees)"
    return math.sin(a * math.pi / 180)


def cos(a):
    "cos (angle in degrees)"
    return math.cos(a * math.pi / 180)


def tan(a):
    "tan (angle in degrees)"
    return math.tan(a * math.pi / 180)


def atan(a, b=None):
    "tan⁻¹ (angle in degrees)"
    return (math.atan(a) if b is None else math.atan2(a, b)) * 180 / math.pi


def stack(*x: Sequence[int | float | Shape]):
    """
    A sequence of 2d things, loft-ed.

    A z1 B z2 C: place shape A at z=0, shape B at z=z1, shape C at z=z1+z2. Loft them.
    A B: complete the object with shape A and start another with shape B.
    A z None: shorthand for A z A.
    """
    obj = Part()
    d = None
    p = 0
    i = iter(x)
    d = last_shape = next(i)
    did_num = False
    for w in i:
        if isinstance(w, (int, float)):
            if did_num:
                raise ValueError("two consecutive numbers?")
            p += w
            did_num = True
        elif did_num:
            if w is None:
                w = last_shape
            else:
                last_shape = w
            nd = Up(p) * w
            obj += loft([d, nd])
            d = nd
            did_num = False
        elif w is None:
            raise ValueError("Real shape wanted")
        else:
            d = Up(p) * w
            last_shape = w
            did_num = False
    if did_num:
        raise ValueError("Trailing number")
    return obj


s_o = ...


def show(obj, name=None, dest=None):
    """
    Display this thing.

    If "dest" is a file or directory, write a STEP file there.
    """
    if name is None:
        name = f"{obj}_{(id(obj) / 99) % 10000}"

    global s_o
    if s_o is Ellipsis:
        import inspect

        f = inspect.currentframe()
        while f is not None and "show_object" not in f.f_globals:
            f = f.f_back
        s_o = None if f is None else f.f_globals["show_object"]
    if s_o is not None:
        s_o(getattr(obj, "wrapped", obj), name)

    if dest is not None:
        dest = Path(dest)
        if dest.is_dir():
            export_stl(obj, str(dest / f"{name}.stl"))
            dest /= f"{name}.step"
        export_step(obj, str(dest))


def read_step(name):
    """
    Read this STEP file.
    """
    return import_step(name)


def read_stl(name):
    """
    Read this STL file.
    """
    return import_stl(name)


def quarter(res):
    """
    Given an object in the +x/+y quarter, mirror it twice.
    """
    res_xz = mirror(res, about=Plane.XZ)
    return res + res_xz + mirror(res_xz + res, about=Plane.YZ)


def gear_turn(n1, a1, n2, a2, minimize=True):
    """
    Given gear1 with n1 teeth, turned to angle a1, and gear2 with n2 teeth positioned at angle a2 relative to gear 1:
    how far do we turn gear2 so that its teeth mesh with gear1?

    This code assumes that 0° means the same thing for both gears (e.g. center of a tooth).

    If @minimize is true (the default), the resulting angle will be between 0 and 360/n2.
    """
    # step 1, assume that gear 2 sits on top of gear 1: turn gear 2 by 180° plus half a tooth.
    res = 180 * (1 + 1 / n2)  # 180+360/n2/2

    # step 2, turn gear 1 counterclockwise by a1, which turns gear 2 clockwise by the gear ratio.
    res += a1 * (n1 / n2)

    # step 3, physically move gear2 around clockwise
    res += a2 * (1 + n1 / n2)

    # step 4, gear 2 is rotationally symmetric
    if minimize:
        res %= 360 / n2

    return res


def pos(x=None, y=None):
    "Line drawing et al. Start position."
    ctx = BuildLine._get_context(log=False)  # noqa:SLF001
    try:
        res = ctx._mt_pos  # noqa:SLF001
    except AttributeError:
        if x is None:
            raise
        res = None
    if y is not None:
        ctx._mt_pos = (x, y)
    return res


def lto(x=None, y=None):
    "Line to absolute position"
    ctx = BuildLine._get_context(log=False)  # noqa:SLF001
    p = ctx._mt_pos  # noqa:SLF001
    if x is None:
        x = p[0]
    if y is None:
        y = p[1]
    p2 = (x, y)
    ctx._mt_pos = p2  # noqa:SLF001
    Line(p, p2)


def li(x=0, y=0):
    "Relative line"
    ctx = BuildLine._get_context(log=False)  # noqa:SLF001
    p = ctx._mt_pos  # noqa:SLF001
    p2 = (p[0] + x, p[1] + y)
    ctx._mt_pos = p2  # noqa:SLF001
    Line(p, p2)


def RX(x):
    "Rotate 90°*x"
    return Rot(90 * x, 0, 0)


RX1 = RX(1)
RX2 = RX(2)
RX3 = RX(3)


def RY(y):
    "Rotate 90°*y"
    return Rot(0, 90 * y, 0)


RY1 = RY(1)
RY2 = RY(2)
RY3 = RY(3)


def RZ(z):
    "Rotate 90°*z"
    return Rot(0, 0, 90 * z)


RZ0 = R0 = RZ(0)
RZ1 = R1 = RZ(1)
RZ2 = R2 = RZ(2)
RZ3 = R3 = RZ(3)

AL = (Align.CENTER, Align.CENTER, Align.MIN)



"""
Miscellaneous solids.
"""

from __future__ import annotations

try:
    import cadquery as cq
except ImportError:
    pass
else:
    __all__ = ["Box", "Cone", "Cylinder", "Loft", "Sphere", "Torus", "Wedge"]

    Box = cq.Solid.makeBox
    Cone = cq.Solid.makeCone
    Cylinder = cq.Solid.makeCylinder
    Loft = cq.Solid.makeLoft
    Sphere = cq.Solid.makeSphere
    Torus = cq.Solid.makeTorus
    Wedge = cq.Solid.makeWedge



"""
This module dies some 3d stuff.
"""

from __future__ import annotations

import anyio

# pylint: disable=missing-module-docstring
import logging
import asyncclick as click
from pathlib import Path as FSPath

from moat.util import load_subgroup

logger = logging.getLogger(__name__)

usage1 = """
"moat cad" collects some useful bits and pieces for 3D editing with Python.
"""


@load_subgroup(sub_pre="moat.link", sub_post="cli", ext_pre="moat.link", ext_post="_main.cli")
@click.pass_context
async def cli(ctx):
    """
    MoaT 3D editor support.

    This collection of commands is useful for managing and building MoaT itself.
    """


@cli.command("edit")
@click.pass_obj
@click.argument("files", nargs=-1, default=None)
async def edit_(obj, files):
    "Start the editor."
    #    base: "/opt/cq"
    # paths:
    #  - "/src/CQ-editor"
    #  - "/src/build123d/src"
    #  - "/src/bd_warehouse/src"

    if len(files) > 1:
        raise click.UsageError("Max one file")
    import moat
    import sys
    import os
    from tempfile import SpooledTemporaryFile

    cfg = obj.cfg
    env = os.environ.copy()

    buf = SpooledTemporaryFile(mode="w+")
    await anyio.run_process(
        [f"{cfg.cad.base}/bin/python3", "-c", "import sys; print(repr(sys.path))"],
        stdout=buf,
    )
    buf.seek(0)
    pypath = (
        cfg.cad.paths
        + eval(buf.read())
        + [str(FSPath(p).parent.absolute()) for p in moat.__path__]
    )
    env["PYTHONPATH"] = os.pathsep.join(pypath)
    await anyio.run_process(
        [f"{cfg.cad.base}/bin/python3", "-mcq_editor"] + list(files),
        env=env,
        # stdin=subprocess.DEVNULL,
        stdout=sys.stdout,
        stderr=sys.stderr,
        # cwd: 'StrOrBytesPath | None' = None,
        # startupinfo: 'Any' = None,
        # creationflags: 'int' = 0,
        # start_new_session: 'bool' = False,
        # pass_fds: 'Sequence[int]' = (),
        # user: 'str | int | None' = None,
        # group: 'str | int | None' = None,
        # extra_groups: 'Iterable[str | int] | None' = None,
        # umask: 'int' = -1
    )


@cli.command("run")
@click.pass_obj
@click.argument("file", nargs=1)
@click.argument("args", nargs=-1)
async def run_(obj, file, args):
    "Run a script."

    import moat
    import sys
    import os
    from tempfile import SpooledTemporaryFile

    cfg = obj.cfg
    env = os.environ.copy()

    buf = SpooledTemporaryFile(mode="w+")
    res = await anyio.run_process(
        [f"{cfg.cad.base}/bin/python3", "-c", "import sys; print(repr(sys.path))"],
        stdout=buf,
    )
    buf.seek(0)
    pypath = (
        cfg.cad.paths
        + eval(buf.read())
        + [str(FSPath(p).parent.absolute()) for p in moat.__path__]
    )
    env["PYTHONPATH"] = os.pathsep.join(pypath)
    res = await anyio.open_process(
        [f"{cfg.cad.base}/bin/python3", file] + list(args),
        env=env,
        stdin=sys.stdin,
        stdout=sys.stdout,
        stderr=sys.stderr,
        # cwd: 'StrOrBytesPath | None' = None,
        # startupinfo: 'Any' = None,
        # creationflags: 'int' = 0,
        # start_new_session: 'bool' = False,
        # pass_fds: 'Sequence[int]' = (),
        # user: 'str | int | None' = None,
        # group: 'str | int | None' = None,
        # extra_groups: 'Iterable[str | int] | None' = None,
        # umask: 'int' = -1
    )
    await res.wait()


@cli.command("gcode")
@click.option("-p","--printer",type=str,default="xl")
async def gcode(printer):
    "Emit gcode template"
    from jinja2 import Environment,PackageLoader
    from anyio import Path

    f=Path(__file__).parent / "_templates" / f"{printer}.gcode.jinja"
    env=Environment(
        keep_trailing_newline=True,
    )
    tmpl=env.from_string(await f.read_text())
    out=tmpl.render()
    print(out)
    




"""
Miscellaneous helpers.
"""

from __future__ import annotations

try:
    import cadquery as cq
except ImportError:
    cq = None

__all__ = ["Slider", "Mount", "Ridge", "WoodScrew"]

from math import tan, pi
import contextlib

with contextlib.suppress(ImportError):
    from .things import Cone


def rad(a):
    return a * pi / 180


def Ridge(length, width, inset=0):
    """
    Returns a triangular profile for slide-ins etc.
    """
    r = cq.Workplane("XY").moveTo(0, 0)
    if inset < 0:
        r = r.line(-inset, 0)
    r = r.line(width, width).line(-width, width)
    if inset < 0:
        r = r.line(inset, 0)
    r = r.close().extrude(length)
    return r


def Slider(x, y, size=2, inset=0, chamfer=None, back=True, centered=True):
    """
    Returns a workplane with a slide-in guide open at -Y, centered on the origin.

    Set "inset" to whatever tolerance you need, for the subtractive part.

    @chamfer is None (sharp end, default), False (flat end) or True (chamfered).
    Set @back to False if you don't want/need the third side.
    """

    cap = -size if chamfer else size if chamfer is None else 0

    # sliders
    def hook(ws, offset, length):
        d = -1 if offset > 0 else 1
        ws = (
            ws.moveTo(offset, 0)
            .line(0, size * 3 + cap + inset)
            .line(d * (size), -cap)
            .line(d * (size + inset), -size - inset)
            .line(-d * size, -size)
            .line(0, -size)
            .close()
            .extrude(length)
        )
        return ws

    h1 = hook(cq.Workplane("XZ"), x, -y)
    h2 = hook(cq.Workplane("XZ"), 0, -y)
    res = h1.union(h2, clean=False)
    if back:
        h3 = hook(cq.Workplane("YZ"), y, x)
        res = res.union(h3, clean=False)
    elif back is None:
        h3 = (
            cq.Workplane("XY")
            .box(x, size * 3, size * 3, centered=False)
            .translate((0, y - size * 3, 0))
        )
        res = res.union(h3)
    if centered:
        res = res.translate((-x / 2, -y / 2, 0))
    return res.clean()


def Mount(length, inner, outer=None, cone=0):
    """A simple ring around a hole"""
    if outer is None:
        outer = inner * 1.2
    ws = cq.Workplane("XY").circle((outer + cone) / 2).extrude(length - cone)
    if cone:
        ws = (
            ws.faces(">Z")
            .workplane()
            .circle((outer + cone) / 2)
            .workplane(offset=cone)
            .circle(2)
            .loft()
        )
    ws = ws.faces(">Z").workplane().circle(inner / 2).cutThruAll()
    return ws


def WoodScrew(height, outer, inner, angle=45, head=None, negate=False):
    """The mount for a wood screw, i.e. one with a non-flat head.

    Hole not included.

    The mount is @height high and has @outer diameter. The screw's diameter
    is @inner and the head's @angle starts at the @head (diameter)'s edge.
    """
    if head is None:
        head = inner * 3 / 2

    h_head = (head - inner) / 2 * tan(rad(angle))

    if negate:
        return (
            cq.Workplane("XY")
            .at(0, 0)
            .circle(inner / 2)
            .extrude(height)
            .add(
                Cone(inner / 2, head / 2, (head - inner) / 2 * tan(rad(angle))).off_z(
                    height - h_head,
                ),
            )
        )

    ws = (
        cq.Workplane("XY")
        .at(0, 0)
        .circle(outer / 2)
        .circle(inner / 2)
        .extrude(height)
        # .faces(">Z").workplane()
        # .circle(head/2)
        .cut(
            Cone(inner / 2, head / 2, (head - inner) / 2 * tan(rad(angle))).off_z(height - h_head),
        )
    )

    return ws



"""
Export helpers.
"""

from __future__ import annotations

try:
    import cadquery as cq
except ImportError:
    cq = None
from pathlib import Path

__all__ = []


if cq is not None:

    def _export(self, filename):
        cq.Assembly(name=Path(filename).stem).add(self).save(
            filename,
            cq.exporters.ExportTypes.STEP,
            mode=cq.exporters.assembly.ExportModes.FUSED,
        )
        return self

    cq.Workplane.export = _export
